"""
Admin ui common utilities.
"""
import warnings
from fnmatch import fnmatch
from importlib import import_module

from django.conf import settings
from django.contrib import admin
from django.urls import reverse


def uniquify(value, seen_values):
    """ Adds value to seen_values set and ensures it is unique """
    id = 1
    new_value = value
    while new_value in seen_values:
        new_value = "%s%s" % (value, id)
        id += 1
    seen_values.add(new_value)
    return new_value


def get_admin_site(context=None, request=None):
    dashboard_cls = getattr(
        settings,
        "ADMIN_TOOLS_INDEX_DASHBOARD",
        "jnt_admin_tools.dashboard.dashboards.DefaultIndexDashboard",
    )

    if isinstance(dashboard_cls, dict):
        if context:
            request = context.get("request")
        curr_url = request.path
        for key in dashboard_cls:
            mod, inst = key.rsplit(".", 1)
            mod = import_module(mod)
            admin_site = getattr(mod, inst)
            admin_url = reverse("%s:index" % admin_site.name)
            if curr_url.startswith(admin_url):
                return admin_site
    else:
        return admin.site
    raise ValueError('Admin site matching "%s" not found' % dashboard_cls)


def get_admin_site_name(context):
    return get_admin_site(context).name


def get_avail_models(request):
    """ Returns (model, perm,) for all models user can possibly see """
    items = []
    admin_site = get_admin_site(request=request)

    for model, model_admin in admin_site._registry.items():
        perms = model_admin.get_model_perms(request)
        if True not in perms.values():
            continue
        items.append((model, perms,))
    return items


def filter_models(request, models, exclude):
    """
    Returns (model, perm,) for all models that match models/exclude patterns
    and are visible by current user.
    """
    items = get_avail_models(request)
    included = []

    def full_name(model):
        return "%s.%s" % (model.__module__, model.__name__)

    # I beleive that that implemented
    # O(len(patterns)*len(matched_patterns)*len(all_models))
    # algorythm is fine for model lists because they are small and admin
    # performance is not a bottleneck. If it is not the case then the code
    # should be optimized.

    if len(models) == 0:
        included = items
    else:
        for pattern in models:
            wildcard_models = []
            for item in items:
                model, perms = item
                model_str = full_name(model)
                if model_str == pattern:
                    # exact match
                    included.append(item)
                elif (
                    fnmatch(model_str, pattern) and item not in wildcard_models
                ):
                    # wildcard match, put item in separate list so it can be
                    # sorted alphabetically later
                    wildcard_models.append(item)
            if wildcard_models:
                # sort wildcard matches alphabetically before adding them
                wildcard_models.sort(
                    key=lambda x: x[0]._meta.verbose_name_plural
                )
                included += wildcard_models

    result = included[:]
    for pattern in exclude:
        for item in included:
            model, perms = item
            if fnmatch(full_name(model), pattern):
                try:
                    result.remove(item)
                except ValueError:  # if the item was already removed skip
                    pass
    return result


class AppListElementMixin:
    """
    Mixin class used by both the AppListDashboardModule and the
    AppListMenuItem (to honor the DRY concept).
    """

    def _visible_models(self, request):
        # compatibility layer: generate models/exclude patterns
        # from include_list/exclude_list args

        if self.include_list:
            warnings.warn(
                "`include_list` is deprecated for ModelList and AppList and "
                "will be removed in future releases. Please use `models` "
                "instead.",
                DeprecationWarning,
            )

        if self.exclude_list:
            warnings.warn(
                "`exclude_list` is deprecated for ModelList and AppList and "
                "will be removed in future releases. Please use `exclude` "
                "instead.",
                DeprecationWarning,
            )

        included = self.models[:]
        included.extend([elem + "*" for elem in self.include_list])

        excluded = self.exclude[:]
        excluded.extend([elem + "*" for elem in self.exclude_list])
        if self.exclude_list and not included:
            included = ["*"]
        return filter_models(request, included, excluded)

    def _get_admin_app_list_url(self, model, context):
        """
        Returns the admin change url.
        """
        app_label = model._meta.app_label
        return reverse(
            f"{get_admin_site_name(context)}:app_list", args=(app_label,)
        )

    def _get_admin_change_url(self, model, context):
        """
        Returns the admin change url.
        """
        app_label = model._meta.app_label
        return reverse(
            f"{get_admin_site_name(context)}:{app_label}_{model.__name__.lower()}_changelist"
        )

    def _get_admin_add_url(self, model, context):
        """
        Returns the admin add url.
        """
        app_label = model._meta.app_label
        return reverse(
            f"{get_admin_site_name(context)}:{app_label}_{model.__name__.lower()}_add"
        )



from typing import Any, Dict, List, Tuple

from django import forms

DEFAULT_PERMISSIONS = ("add", "change", "delete", "view")


class PermissionSelectMultipleWidget(forms.CheckboxSelectMultiple):
    """
    A Widget for displaying all permissions in system.

    forms.py::

        from jnt_admin_tools.widgets import PermissionSelectMultipleWidget
        from django import forms
        from django.contrib.auth.models import Group

        class GroupAdminForm(forms.ModelForm):
            class Meta:
                model = Group
                fields = '__all__'
                widgets = {
                    'permissions': PermissionSelectMultipleWidget
                }

    admin.py::

        from django.contrib import admin
        from django.contrib.auth.models import Group
        from test_app.forms import GroupAdminForm

        @admin.register(Group)
        class GroupAdmin(admin.ModelAdmin):
            form = GroupAdminForm

    .. image:: images/widgets/permission_select_multiple_widget.png
    """

    template_name = "jnt_admin_tools/widgets/permissions.html"
    custom_permission_types: List[str] = []
    groups_permissions: List[str] = []

    def get_context(self, name, value, attrs):
        if value is None:
            value = []

        return {
            "name": name,
            "value": value,
            "table": self.get_table(),
            "groups_permissions": self.groups_permissions,
            "default_permission_types": DEFAULT_PERMISSIONS,
            "custom_permission_types": self.custom_permission_types,
        }

    def get_table(self):
        table = []
        row: Dict[str, Any] = {}

        for permission in self.choices.queryset.select_related(
            "content_type"
        ).all():
            row, created = self._update_or_create_permission_row(
                permission, row
            )
            if created:
                table.append(row)

        return table

    def _update_or_create_permission_row(
        self, permission, last_row: Dict[str, Any]
    ) -> Tuple[Dict[str, Any], bool]:
        codename = permission.codename
        model_part = "_" + permission.content_type.model
        permission_type = codename
        if permission_type.endswith(model_part):
            permission_type = permission_type[: -len(model_part)]

        app = permission.content_type.app_label.replace("_", " ")
        model_class = permission.content_type.model_class()
        model_verbose_name = (
            model_class._meta.verbose_name if model_class else None
        )

        if (
            permission_type
            not in list(DEFAULT_PERMISSIONS) + self.custom_permission_types
        ):
            self.custom_permission_types.append(permission_type)

        is_app_or_model_different = not last_row or (
            last_row["model_class"] != model_class or last_row["app"] != app
        )
        created = False

        if is_app_or_model_different:
            created = True
            row = dict(
                model=model_verbose_name,
                model_class=model_class,
                app=app,
                permissions={},
            )
        else:
            row = last_row

        row["permissions"][permission_type] = permission

        return row, created

    class Media:
        css = {"all": ("jnt_admin_tools/css/widgets/permissions.css",)}



from django.conf.urls import url
from jnt_admin_tools.dashboard import views

urlpatterns = [
    url(
        r"^set_preferences/(?P<dashboard_id>.+)/$",
        views.set_preferences,
        name="admin-tools-dashboard-set-preferences",
    ),
]



from functools import reduce, wraps

from jnt_admin_tools.services.urls import (
    admin_change_url,
    admin_changelist_url,
)
from django.forms.forms import pretty_name
from django.utils.html import format_html


def getattr_nested(obj, attr):
    return reduce(getattr, [obj] + attr.split("__"))


def admin_link(attr, short_description=None, empty_description="-"):
    """Decorator used for rendering a link to a related model in
    the admin detail page.

    ``attr (str)``
        Name of the related field.
    ``short_description (str)``
        Name if the field.
        Default value: None.
    ``empty_description (str)``
        Value to display if the related field is None.
        Default value: -.

    The wrapped method receives the related object and should
    return the link text.

    Usage::

        from django.contrib import admin
        from jnt_admin_tools.decorators import admin_link
        from test_app.models import Foo

        @admin.register(Foo)
        class FooAdmin(admin.ModelAdmin):
            fields = ('name', 'bar', 'bar_link')
            readonly_fields = ('bar_link',)

            @admin_link('bar')
            def bar_link(self, obj):
                return obj.name

    .. image:: images/decorators/decorator_admin_link.png
    """

    def wrap(func):
        @wraps(func)
        def field_func(self, obj):
            related_obj = getattr_nested(obj, attr)
            if related_obj is None:
                return empty_description
            url = admin_change_url(related_obj)
            return format_html(
                '<a href="{0}">{1}</a>', url, func(self, related_obj)
            )

        field_func.short_description = (
            short_description if short_description else pretty_name(attr)
        )
        field_func.allow_tags = True
        return field_func

    return wrap


def admin_changelist_link(
    attr, short_description=None, empty_description="-", query_string=None
):
    """Decorator used for rendering a link to the list display of
    a related model in the admin changelist page.

    ``attr (str)``
        Name of the related field.
    ``short_description (str)``
        Field display name.
        Default value: None.
    ``empty_description (str)``
        Value to display if the related field is None.
        Default value: -.
    ``query_string (function)``
        Optional callback for adding a query string to the link.
        Receives the object and should return a query string.
        Default value: None.

    The wrapped method receives the related object and
    should return the link text.

    Usage::

        from jnt_admin_tools.decorators import admin_changelist_link
        from django.contrib import admin
        from test_app.models import Bar

        @admin.register(Bar)
        class BarAdmin(admin.ModelAdmin):
            fields = ('name', 'foos', 'custom_field')
            readonly_fields = ('foos', 'custom_field')

            @admin_changelist_link('foos', query_string=lambda bar: f'bar_id={bar.pk}')
            def foos(self, foos):
                return ', '.join(str(foo) for foo in foos.all())

    .. image:: images/decorators/decorator_admin_changelist_link.png
    """

    def wrap(func):
        @wraps(func)
        def field_func(self, obj):
            related_obj = getattr_nested(obj, attr)
            if related_obj is None:
                return empty_description
            url = admin_changelist_url(related_obj.model)
            if query_string:
                url += "?" + query_string(obj)
            return format_html(
                '<a href="{0}">{1}</a>', url, func(self, related_obj)
            )

        field_func.short_description = (
            short_description if short_description else pretty_name(attr)
        )
        field_func.allow_tags = True
        return field_func

    return wrap


def admin_field(short_description=None, allow_tags=True):
    """Decorator used for custom field or add link in
    the admin detail page.

    ``short_description (str)``
        Description of the field.
    ``allow_tags (bool)``
        Allow tags.

    The wrapped method receives the custom field.

    Usage::

        from jnt_admin_tools.decorators import admin_field
        from django.contrib import admin
        from django.utils.safestring import mark_safe
        from test_app.models import Bar

        @admin.register(Bar)
        class BarAdmin(admin.ModelAdmin):
            fields = ('name', 'foos', 'custom_field')
            readonly_fields = ('foos', 'custom_field')

            @admin_field('Custom field')
            def custom_field(self, obj):
                return mark_safe('<h1>custom field</h1>')

    .. image:: images/decorators/decorator_admin_field.png
    """

    def wrap(func):
        @wraps(func)
        def field_func(self, obj):
            return func(self, obj)

        field_func.short_description = short_description
        field_func.allow_tags = allow_tags
        return field_func

    return wrap



from jnt_admin_tools.services.urls import admin_autocomplete_url
from django.contrib.admin.views.autocomplete import AutocompleteJsonView
from django.http import Http404, JsonResponse
from django.utils.text import capfirst


class ContentTypeAutocompleteView(AutocompleteJsonView):
    # copied from base class
    def get(self, request, *args, **kwargs):
        if not self.model_admin.get_search_fields(request):
            raise Http404(
                "%s must have search_fields for the autocomplete_view."
                % type(self.model_admin).__name__
            )
        if not self.has_perm(request):
            return JsonResponse({"error": "403 Forbidden"}, status=403)

        self.term = request.GET.get("term", "")
        self.paginator_class = self.model_admin.paginator
        self.object_list = self.get_queryset()
        context = self.get_context_data()
        return JsonResponse(
            {
                "results": [
                    {
                        "id": str(obj.pk),
                        "text": capfirst(obj.model),
                        # modify
                        "autocompleteUrl": self.get_autocomplete_ul(obj),
                    }
                    for obj in context["object_list"]
                ],
                "pagination": {"more": context["page_obj"].has_next()},
            }
        )

    def get_queryset(self):
        queryset = super().get_queryset()

        ids = self.request.GET.get("ids")

        if ids:
            queryset = queryset.filter(id__in=[int(x) for x in ids.split(",")])

        return queryset

    def get_autocomplete_ul(self, content_type):
        return admin_autocomplete_url(content_type.model_class())



from django.core.management import BaseCommand

from jnt_admin_tools.dashboard.models import DashboardPreferences


class Command(BaseCommand):
    help = """Clears dashboard preferences."""

    def handle(self, *args, **options):
        DashboardPreferences._default_manager.all().delete()



from django.contrib.admin.widgets import AutocompleteSelect
from django import forms
from django.contrib import admin
from django.conf import settings
from django.contrib.admin.widgets import SELECT2_TRANSLATIONS, get_language
from django.db.models.fields.related_descriptors import ManyToManyDescriptor
from django.forms.widgets import Media, MEDIA_TYPES


class AutocompleteFilter(admin.SimpleListFilter):
    template = "jnt_admin_tools/autocomplete_filter/autocomplete_filter.html"
    title = ""
    field_name = ""
    is_placeholder_title = False
    widget_attrs = {}
    rel_model = None

    class Media:
        extra = "" if settings.DEBUG else ".min"
        i18n_name = SELECT2_TRANSLATIONS.get(get_language())
        i18n_file = (
            ("admin/js/vendor/select2/i18n/%s.js" % i18n_name,)
            if i18n_name
            else ()
        )
        js = (
            (
                "admin/js/vendor/jquery/jquery%s.js" % extra,
                "admin/js/vendor/select2/select2.full%s.js" % extra,
            )
            + i18n_file
            + (
                "admin/js/jquery.init.js",
                "admin/js/autocomplete.js",
                "jnt_admin_tools/js/autocomplete_filter/autocomplete_filter_qs.js",
            )
        )
        css = {
            "screen": (
                "admin/css/vendor/select2/select2%s.css" % extra,
                "admin/css/autocomplete.css",
                "jnt_admin_tools/css/autocomplete_filter/autocomplete_fix.css",
            ),
        }

    def __init__(self, request, params, model, model_admin):
        if self.parameter_name:
            raise AttributeError(
                "Rename attribute 'parameter_name' to "
                "'field_name' for {0}".format(self.__class__)
            )
        self.parameter_name = "{0}__id__exact".format(self.field_name)
        super().__init__(request, params, model, model_admin)

        if self.rel_model:
            model = self.rel_model

        remote_field = model._meta.get_field(self.field_name).remote_field

        attrs = self.widget_attrs.copy()

        field = forms.ModelChoiceField(
            queryset=self.get_field_queryset(model),
            widget=AutocompleteSelect(remote_field, model_admin.admin_site),
            required=False,
        )

        self._add_media(model_admin)

        attrs["id"] = "id-{0}-dal-filter".format(self.field_name)

        if self.is_placeholder_title:
            attrs["data-placeholder"] = "By " + self.title

        self.rendered_widget = field.widget.render(
            name=self.parameter_name,
            value=self.used_parameters.get(self.parameter_name, ""),
            attrs=attrs,
        )

    def _add_media(self, model_admin):
        if not hasattr(model_admin, "Media"):
            raise Exception(
                "Please add empty Media class to %s." % model_admin
            )

        def _get_media(obj):
            return Media(media=getattr(obj, "Media", None))

        media = (
            _get_media(model_admin)
            + _get_media(AutocompleteFilter)
            + _get_media(self)
        )

        for name in MEDIA_TYPES:
            setattr(model_admin.Media, name, getattr(media, "_" + name))

    def has_output(self):
        return True

    def lookups(self, request, model_admin):
        return ()

    def queryset(self, request, queryset):
        if self.value():
            return queryset.filter(**{self.parameter_name: self.value()})
        else:
            return queryset

    def get_field_queryset(self, model):
        field = getattr(model, self.field_name)
        if isinstance(field, ManyToManyDescriptor):
            return field.rel.model.objects.all()

        return field.get_queryset()



from django.contrib.admin.views.decorators import staff_member_required
from django.http import HttpResponse, HttpResponseRedirect
from django.shortcuts import get_object_or_404, render
from django.contrib import messages

try:
    from django.views.decorators.csrf import csrf_exempt
except ImportError:
    from django.contrib.csrf.middleware import csrf_exempt

from .forms import BookmarkForm
from .models import Bookmark


@staff_member_required
@csrf_exempt
def add_bookmark(request):
    """
    This view serves and validates a bookmark form.
    If requested via ajax it also returns the drop bookmark form to replace the
    add bookmark form.
    """
    if request.method == "POST":
        form = BookmarkForm(user=request.user, data=request.POST)
        if form.is_valid():
            bookmark = form.save()
            if not request.is_ajax():
                messages.success(request, "Bookmark added")
                if request.POST.get("next"):
                    return HttpResponseRedirect(request.POST.get("next"))
                return HttpResponse("Added")
            return render(
                request,
                "jnt_admin_tools/menu/remove_bookmark_form.html",
                {"bookmark": bookmark, "url": bookmark.url},
            )
    else:
        form = BookmarkForm(user=request.user)
    return render(
        request,
        "jnt_admin_tools/menu/form.html",
        {"form": form, "title": "Add Bookmark"},
    )


@staff_member_required
@csrf_exempt
def edit_bookmark(request, id):
    bookmark = get_object_or_404(Bookmark, id=id)
    if request.method == "POST":
        form = BookmarkForm(
            user=request.user, data=request.POST, instance=bookmark
        )
        if form.is_valid():
            form.save()
            if not request.is_ajax():
                messages.success(request, "Bookmark updated")
                if request.POST.get("next"):
                    return HttpResponseRedirect(request.POST.get("next"))
            return HttpResponse("Saved")
    else:
        form = BookmarkForm(user=request.user, instance=bookmark)
    return render_to_response(
        "jnt_admin_tools/menu/form.html",
        {"form": form, "title": "Edit Bookmark"},
    )


@staff_member_required
@csrf_exempt
def remove_bookmark(request, id):
    """
    This view deletes a bookmark.
    If requested via ajax it also returns the add bookmark form to replace the
    drop bookmark form.
    """
    bookmark = get_object_or_404(Bookmark, id=id, user=request.user)
    if request.method == "POST":
        bookmark.delete()
        if not request.is_ajax():
            messages.success(request, "Bookmark removed")
            if request.POST.get("next"):
                return HttpResponseRedirect(request.POST.get("next"))
            return HttpResponse("Deleted")
        return render(
            request,
            "jnt_admin_tools/menu/add_bookmark_form.html",
            {
                "url": request.POST.get("next"),
                "title": "**title**",  # replaced on the javascript side
            },
        )
    return render(
        request,
        "jnt_admin_tools/menu/delete_confirm.html",
        {"bookmark": bookmark, "title": "Delete Bookmark"},
    )



from jnt_admin_tools.utils import AppListElementMixin
from django.apps import apps as django_apps
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _


class MenuItem:
    """
    This is the base class for custom menu items.
    A menu item can have the following properties:

    ``title``
        String that contains the menu item title, make sure you use the
        django gettext functions if your application is multilingual.
        Default value: 'Untitled menu item'.

    ``url``
        String that contains the menu item URL.
        Default value: '#'.

    ``css_classes``
        A list of css classes to be added to the menu item ``li`` class
        attribute. Default value: [].

    ``accesskey``
        The menu item accesskey. Default value: None.

    ``description``
        An optional string that will be used as the ``title`` attribute of
        the menu-item ``a`` tag. Default value: None.

    ``enabled``
        Boolean that determines whether the menu item is enabled or not.
        Disabled items are displayed but are not clickable.
        Default value: True.

    ``template``
        The template to use to render the menu item.
        Default value: 'jnt_admin_tools/menu/item.html'.

    ``children``
        A list of children menu items. All children items must be instances of
        the ``MenuItem`` class.
    """

    title = "Untitled menu item"
    url = "#"
    css_classes = None
    accesskey = None
    description = None
    enabled = True
    template = "jnt_admin_tools/menu/item.html"
    children = None

    def __init__(self, title=None, url=None, **kwargs):

        if title is not None:
            self.title = title

        if url is not None:
            self.url = url

        for key in kwargs:
            if hasattr(self.__class__, key):
                setattr(self, key, kwargs[key])
        self.children = self.children or []
        self.css_classes = self.css_classes or []

    def init_with_context(self, context):
        """
        Like for menus, menu items have a ``init_with_context`` method that is
        called with a ``django.template.RequestContext`` instance as unique
        argument.
        This gives you enough flexibility to build complex items, for example,
        let's build a "history" menu item, that will list the last ten visited
        pages::

            from jnt_admin_tools.menu.items import MenuItem

            class HistoryMenuItem(MenuItem):
                title = 'History'

                def init_with_context(self, context):
                    request = context['request']
                    # we use sessions to store the visited pages stack
                    history = request.session.get('history', [])
                    for item in history:
                        self.children.append(MenuItem(
                            title=item['title'],
                            url=item['url']
                        ))
                    # add the current page to the history
                    history.insert(0, {
                        'title': context['title'],
                        'url': request.META['PATH_INFO']
                    })
                    if len(history) > 10:
                        history = history[:10]
                    request.session['history'] = history

        Here's a screenshot of our history item:

        .. image:: images/history_menu_item.png
        """
        pass

    def is_selected(self, request):
        """
        Helper method that returns ``True`` if the menu item is active.
        A menu item is considered as active if it's URL or one of its
        descendants URL is equals to the current URL.
        """
        current_url = request.get_full_path()
        return (
            self.url == current_url
            or len([c for c in self.children if c.is_selected(request)]) > 0
        )

    def is_empty(self):
        """
        Helper method that returns ``True`` if the menu item is empty.
        This method always returns ``False`` for basic items, but can return
        ``True`` if the item is an AppList.
        """
        return False


class AppList(MenuItem, AppListElementMixin):
    """
    A menu item that lists installed apps an their models.
    In addition to the parent :class:`~jnt_admin_tools.menu.items.MenuItem`
    properties, the ``AppList`` has two extra properties:

    ``models``
        A list of models to include, only models whose name (e.g.
        "blog.comments.Comment") match one of the strings (e.g. "blog.*")
        in the models list will appear in the menu item.

    ``exclude``
        A list of models to exclude, if a model name (e.g.
        "blog.comments.Comment") match an element of this list (e.g.
        "blog.comments.*") it won't appear in the menu item.


    If no models/exclude list is provided, **all apps** are shown.

    Here's a small example of building an app list menu item::

        from jnt_admin_tools.menu import items, Menu

        class MyMenu(Menu):
            def __init__(self, **kwargs):
                super(MyMenu, self).__init__(**kwargs)
                self.children.append(items.AppList(
                    title='Applications',
                    exclude_list=('django.contrib',)
                )

    The screenshot of what this code produces:

    .. image:: images/applist_menu_item.png

    .. note::

        Note that this menu takes into account user permissions, as a
        consequence, if a user has no rights to change or add a ``Group`` for
        example, the ``django.contrib.auth.Group model`` child item won't be
        displayed in the menu.
    """

    def __init__(self, title=None, **kwargs):
        """
        ``AppListMenuItem`` constructor.
        """
        self.models = list(kwargs.pop("models", []))
        self.exclude = list(kwargs.pop("exclude", []))
        self.include_list = kwargs.pop("include_list", [])  # deprecated
        self.exclude_list = kwargs.pop("exclude_list", [])  # deprecated
        super(AppList, self).__init__(title, **kwargs)

    def init_with_context(self, context):
        """
        Please refer to
        :meth:`~jnt_admin_tools.menu.items.MenuItem.init_with_context`
        documentation from :class:`~jnt_admin_tools.menu.items.MenuItem` class.
        """
        items = self._visible_models(context["request"])
        apps = {}
        for model, perms in items:
            if not (perms["change"] or perms.get("view", False)):
                continue
            app_label = model._meta.app_label
            if app_label not in apps:
                apps[app_label] = {
                    "title": django_apps.get_app_config(
                        app_label
                    ).verbose_name,
                    "url": self._get_admin_app_list_url(model, context),
                    "models": [],
                }
            apps[app_label]["models"].append(
                {
                    "title": model._meta.verbose_name_plural,
                    "url": self._get_admin_change_url(model, context),
                }
            )

        for app in sorted(apps.keys()):
            app_dict = apps[app]
            item = MenuItem(title=app_dict["title"], url=app_dict["url"])
            # sort model list alphabetically
            apps[app]["models"].sort(key=lambda x: x["title"])
            for model_dict in apps[app]["models"]:
                item.children.append(MenuItem(**model_dict))
            self.children.append(item)

    def is_empty(self):
        """
        Helper method that returns ``True`` if the applist menu item has no
        children.

        >>> from jnt_admin_tools.menu.items import MenuItem, AppList
        >>> item = AppList(title='My menu item')
        >>> item.is_empty()
        True
        >>> item.children.append(MenuItem(title='foo'))
        >>> item.is_empty()
        False
        >>> item.children = []
        >>> item.is_empty()
        True
        """
        return len(self.children) == 0


class ModelList(MenuItem, AppListElementMixin):
    """
    A menu item that lists a set of models.
    In addition to the parent :class:`~jnt_admin_tools.menu.items.MenuItem`
    properties, the ``ModelList`` has two extra properties:

    ``models``
        A list of models to include, only models whose name (e.g.
        "blog.comments.Comment") match one of the strings (e.g. "blog.*")
        in the include list will appear in the dashboard module.

    ``exclude``
        A list of models to exclude, if a model name (e.g.
        "blog.comments.Comment" match an element of this list (e.g.
        "blog.comments.*") it won't appear in the dashboard module.

    Here's a small example of building a model list menu item::

        from jnt_admin_tools.menu import items, Menu

        class MyMenu(Menu):
            def __init__(self, **kwargs):
                super(MyMenu, self).__init__(**kwargs)
                self.children += [
                    items.ModelList(
                        'Authentication', ['django.contrib.auth.*',]
                    )
                ]

    .. note::

        Note that this menu takes into account user permissions, as a
        consequence, if a user has no rights to change or add a ``Group`` for
        example, the ``django.contrib.auth.Group model`` item won't be
        displayed in the menu.
    """

    def __init__(self, title=None, models=None, exclude=None, **kwargs):
        """
        ``ModelList`` constructor.
        """
        self.models = list(models or [])
        self.exclude = list(exclude or [])
        self.include_list = kwargs.pop("include_list", [])  # deprecated
        self.exclude_list = kwargs.pop("exclude_list", [])  # deprecated

        super(ModelList, self).__init__(title, **kwargs)

    def init_with_context(self, context):
        """
        Please refer to
        :meth:`~jnt_admin_tools.menu.items.MenuItem.init_with_context`
        documentation from :class:`~jnt_admin_tools.menu.items.MenuItem` class.
        """
        items = self._visible_models(context["request"])
        for model, perms in items:
            if not (perms["change"] or perms.get("view", False)):
                continue
            title = model._meta.verbose_name_plural
            url = self._get_admin_change_url(model, context)
            item = MenuItem(title=title, url=url)
            self.children.append(item)

    def is_empty(self):
        """
        Helper method that returns ``True`` if the modellist menu item has no
        children.

        >>> from jnt_admin_tools.menu.items import MenuItem, ModelList
        >>> item = ModelList(title='My menu item')
        >>> item.is_empty()
        True
        >>> item.children.append(MenuItem(title='foo'))
        >>> item.is_empty()
        False
        >>> item.children = []
        >>> item.is_empty()
        True
        """
        return len(self.children) == 0


class Bookmarks(MenuItem, AppListElementMixin):
    """
    A menu item that lists pages bookmarked by the user. This menu item also
    adds an extra button to the menu that allows the user to bookmark or
    un-bookmark the current page.

    Here's a small example of adding a bookmark menu item::

        from jnt_admin_tools.menu import items, Menu

        class MyMenu(Menu):
            def __init__(self, **kwargs):
                super(MyMenu, self).__init__(**kwargs)
                self.children.append(items.Bookmarks('My bookmarks'))

    """

    title = _("Bookmarks")

    def __init__(self, title=None, **kwargs):
        super(Bookmarks, self).__init__(title, **kwargs)
        if "bookmark" not in self.css_classes:
            self.css_classes.append("bookmark")

    def init_with_context(self, context):
        """
        Please refer to
        :meth:`~jnt_admin_tools.menu.items.MenuItem.init_with_context`
        documentation from :class:`~jnt_admin_tools.menu.items.MenuItem` class.
        """
        from jnt_admin_tools.menu.models import Bookmark

        for b in Bookmark.objects.filter(user=context["request"].user):
            self.children.append(MenuItem(mark_safe(b.title), b.url))

        if not len(self.children):
            self.enabled = False

    def is_selected(self, request):
        """
        A bookmark menu item is never considered as active, the real item is.
        """
        return False



"""
Module where admin tools menu classes are defined.
"""

try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse
from django.utils.translation import ugettext_lazy as _

from jnt_admin_tools.menu import items
from jnt_admin_tools.utils import get_admin_site_name


class Menu(object):
    """
    This is the base class for creating custom navigation menus.
    A menu can have the following properties:

    ``template``
        A string representing the path to template to use to render the menu.
        As for any other template, the path must be relative to one of the
        directories of your ``TEMPLATE_DIRS`` setting.
        Default value: "jnt_admin_tools/menu/menu.html".

    ``children``
        A list of children menu items. All children items mus be instances of
        the :class:`~jnt_admin_tools.menu.items.MenuItem` class.

    If you want to customize the look of your menu and it's menu items, you
    can declare css stylesheets and/or javascript files to include when
    rendering the menu, for example::

        from jnt_admin_tools.menu import Menu

        class MyMenu(Menu):
            class Media:
                css = {'all': ('css/mymenu.css',)}
                js = ('js/mymenu.js',)

    Here's a concrete example of a custom menu::

        from django.core.urlresolvers import reverse
        from jnt_admin_tools.menu import items, Menu

        class MyMenu(Menu):
            def __init__(self, **kwargs):
                super(MyMenu, self).__init__(**kwargs)
                self.children += [
                    items.MenuItem('Home', reverse('admin:index')),
                    items.AppList('Applications'),
                    items.MenuItem('Multi level menu item',
                        children=[
                            items.MenuItem('Child 1', '/foo/'),
                            items.MenuItem('Child 2', '/bar/'),
                        ]
                    ),
                ]

    Below is a screenshot of the resulting menu:

    .. image:: images/menu_example.png
    """

    template = "jnt_admin_tools/menu/menu.html"
    children = None

    class Media:
        css = ()
        js = ()

    def __init__(self, **kwargs):
        for key in kwargs:
            if hasattr(self.__class__, key):
                setattr(self, key, kwargs[key])
        self.children = kwargs.get("children", [])

    def init_with_context(self, context):
        """
        Sometimes you may need to access context or request variables to build
        your menu, this is what the ``init_with_context()`` method is for.
        This method is called just before the display with a
        ``django.template.RequestContext`` as unique argument, so you can
        access to all context variables and to the ``django.http.HttpRequest``.
        """
        pass


class DefaultMenu(Menu):
    """
    The default menu displayed by django-admin-tools.
    To change the default menu you'll have to type the following from the
    commandline in your project root directory::

        python manage.py custommenu

    And then set the ``ADMIN_TOOLS_MENU`` settings variable to point to your
    custom menu class.
    """

    def init_with_context(self, context):
        site_name = get_admin_site_name(context)

        self.children += [
            items.MenuItem(_("Dashboard"), reverse("%s:index" % site_name)),
            items.Bookmarks(),
            items.AppList(_("Applications"), exclude=("django.contrib.*",)),
            items.AppList(_("Administration"), models=("django.contrib.*",)),
        ]



from jnt_admin_tools.views import ContentTypeAutocompleteView
from django.contrib import admin
from django.contrib.contenttypes.models import ContentType


@admin.register(ContentType)
class ContentTypeAdmin(admin.ModelAdmin):
    search_fields = ("model",)

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_view_or_change_permission(self, request, obj=None):
        return False

    def has_view_permission(self, request, obj=None):
        return True

    def get_model_perms(self, request):
        return {}

    def autocomplete_view(self, request):
        return ContentTypeAutocompleteView.as_view(model_admin=self)(request)



from django.conf.urls import url
from jnt_admin_tools.menu import views

urlpatterns = [
    url(
        r"^add_bookmark/$",
        views.add_bookmark,
        name="admin-tools-menu-add-bookmark",
    ),
    url(
        r"^edit_bookmark/(?P<id>.+)/$",
        views.edit_bookmark,
        name="admin-tools-menu-edit-bookmark",
    ),
    url(
        r"^remove_bookmark/(?P<id>.+)/$",
        views.remove_bookmark,
        name="admin-tools-menu-remove-bookmark",
    ),
]
